# copyright:
# ----------------------------------------------------------------------------
# This confidential and proprietary software may be used only as authorized
# by a licensing agreement from ARM Limited.
#      (C) COPYRIGHT 2017-2018 ARM Limited, ALL RIGHTS RESERVED
# The entire notice above must be reproduced on all authorized copies and
# copies may only be made to the extent permitted by a licensing agreement
# from ARM Limited.
# ----------------------------------------------------------------------------

import os
import config_system.lex

class LexWrapper:
    def __init__(self, ignore_missing):
        self.lexers = []
        self.root_dir = None
        self.ignore_missing = ignore_missing

    def source(self, fname):
        if self.root_dir is None:
            self.root_dir = os.path.dirname(fname)
        else:
            # All paths are relative to the dir containing the first Mconfig
            fname = os.path.join(self.root_dir, fname)

        if not os.path.exists(fname) and self.ignore_missing:
           return

        with open(fname, "rt") as fp:
            file_contents = fp.read()

        lexer = config_system.lex.create_mconfig_lexer(fname)

        self.push_lexer(lexer)
        self.input(file_contents)

    def current_lexer(self):
        return self.lexers[-1]

    def push_lexer(self, lexer):
        self.lexers.append(lexer)

    def pop_lexer(self):
        self.lexers = self.lexers[0:-1]

    def input(self, input):
        assert self.lexers

        self.current_lexer().input(input)

    def token(self):
        if not self.lexers:
            return None

        t = self.current_lexer().token()

        if t is None:
            self.pop_lexer()
            t = self.token()

        return t
