# Copyright:
# ----------------------------------------------------------------------------
# This confidential and proprietary software may be used only as authorized
# by a licensing agreement from ARM Limited.
#      (C) COPYRIGHT 2018 ARM Limited, ALL RIGHTS RESERVED
# The entire notice above must be reproduced on all authorized copies and
# copies may only be made to the extent permitted by a licensing agreement
# from ARM Limited.
# ----------------------------------------------------------------------------
import logging
from logging.handlers import BufferingHandler

# Like MemoryHandler, but only flush on close
class InfBufferHandler(BufferingHandler):
    __slots__ = ('target', 'buffer')

    def __init__(self, capacity, target):
        super(BufferingHandler, self).__init__(capacity)
        self.setLevel(logging.NOTSET)
        self.target = target
        self.buffer = []

    def shouldFlush(self, record):
        return False

    def flush(self):
        while self.buffer:
            rec = self.buffer.pop(0)
            self.target.emit(rec)

    def close(self):
        self.flush()
        self.target = None
        self.buffer = []

# Count the number of each error type
class ErrorCounterHandler(logging.Handler):
    __slots__ = ('counts')

    def __init__(self, *args, **kwargs):
        super(ErrorCounterHandler, self).__init__(*args, **kwargs)
        self.counts = {
            'NOTSETS': 0,
            'DEBUG': 0,
            'INFO': 0,
            'WARNING': 0,
            'ERROR': 0,
            'CRITICAL': 0,
        }

    def emit(self, record):
        l = record.levelname
        self.counts[l] += 1

    def debugs(self):
        return self.counts['DEBUG']

    def infos(self):
        return self.counts['INFO']

    def warnings(self):
        return self.counts['WARNING']

    def errors(self):
        return self.counts['ERROR']

    def criticals(self):
        return self.counts['CRITICAL']
