/*
 * Copyright:
 * ----------------------------------------------------------------------------
 * This confidential and proprietary software may be used only as authorized
 * by a licensing agreement from ARM Limited.
 *      (C) COPYRIGHT 2017-2018 ARM Limited, ALL RIGHTS RESERVED
 * The entire notice above must be reproduced on all authorized copies and
 * copies may only be made to the extent permitted by a licensing agreement
 * from ARM Limited.
 * ----------------------------------------------------------------------------
 */

package core

import (
	"github.com/google/blueprint"
)

type Defaults struct {
	Library
}

func (m *Defaults) supportedVariants() []string {
	return []string{HOST_TAG, TARGET_TAG}
}

func (m *Defaults) disable() {
	panic("disable() called on Default")
}

func (m *Defaults) setVariant(variant string) {
	m.Library.setVariant(variant)
}

func (m *Defaults) getSplittableProps() *SplittableProps {
	return m.Library.getSplittableProps()
}

func (m *Defaults) GenerateBuildActions(ctx blueprint.ModuleContext) {
}

func DefaultsFactory(config *Config) (blueprint.Module, []interface{}) {
	module := &Defaults{}
	return module.LibraryFactory(config, module)
}

var defaultDepTag = DependencyTag{name: "default"}

type Defaultable interface {
	Build() *Build
	Features() *Features
	Defaults() []string
}

func DefaultDepsMutator(mctx blueprint.BottomUpMutatorContext) {
	if l, ok := mctx.Module().(Defaultable); ok {
		mctx.AddDependency(mctx.Module(), defaultDepTag, l.Defaults()...)
	}
	if gsc, ok := GetGenerateCommon(mctx.Module()); ok {
		mctx.AddDependency(mctx.Module(), defaultDepTag, gsc.Properties.Flag_defaults...)
	}
}
