#!/bin/bash
#
# Copyright:
# ----------------------------------------------------------------------------
# This confidential and proprietary software may be used only as authorized
# by a licensing agreement from ARM Limited.
#      (C) COPYRIGHT 2018 ARM Limited, ALL RIGHTS RESERVED
# The entire notice above must be reproduced on all authorized copies and
# copies may only be made to the extent permitted by a licensing agreement
# from ARM Limited.
# ----------------------------------------------------------------------------
#

# Functions for manipulating paths

# Counts the number of path elements
function count_path_elems() {
    P=$1
    IFS='/'
    set -f $P
    echo $#
}

# Choose the shortest path of the 2 arguments.
# This is expected to be used with equivalent relative and absolute paths.
# Where they are the same length, the first is preferred.
function shortest_path() {
    COUNT1=$(count_path_elems $1)
    COUNT2=$(count_path_elems $2)
    if [ ${COUNT1} -le ${COUNT2} ]; then
        echo ${1}
    else
        echo ${2}
    fi
}

# Return a path that references $2 from $1
# $1 and $2 must exist
# This is a simple implementation. We rely on readlink to sort out symlink issues for us.
# If there are fewer path elements in the absolute version, return that instead.
function relative_path() {
    SRC_ABS=$(readlink -f $1)
    TGT_ABS=$(readlink -f $2)

    BACK=
    # ${TGT_ABS#SRC_ABS} removes the SRC_ABS from TGT_ABS
    if [ "${TGT_ABS#${SRC_ABS}}" != "${TGT_ABS}" ]; then
        # SRC_ABS is a parent of TGT_ABS

        RESULT=${TGT_ABS#${SRC_ABS}/}

    elif [ "${SRC_ABS#${TGT_ABS}}" != "${SRC_ABS}" ]; then
        # TGT_ABS is a parent of SRC_ABS

        while [ "${TGT_ABS}" != "${SRC_ABS}" ]; do
            SRC_ABS=$(dirname ${SRC_ABS})
            BACK="../${BACK}"
        done

        RESULT=${BACK%/}

    else
        CMN_PFX=${SRC_ABS}

        # ${TGT_ABS#CMN_PFX} removes the CMN_PFX from TGT_ABS
        # When we have the correct prefix, the condition will fail
        while [ "${TGT_ABS#${CMN_PFX}}" = "${TGT_ABS}" ]; do
            CMN_PFX=$(dirname ${CMN_PFX})
            BACK="../${BACK}"
        done

        RESULT=${BACK}${TGT_ABS#${CMN_PFX}/}
    fi

    echo ${RESULT}
}
