#!/bin/bash
#
# ----------------------------------------------------------------------------
# This confidential and proprietary software may be used only as
# authorised by a licensing agreement from ARM Limited.
# (C) COPYRIGHT 2017 ARM Limited
# ALL RIGHTS RESERVED
# The entire notice above must be reproduced on all authorised
# copies and copies may only be made to the extent permitted
# by a licensing agreement from ARM Limited.
# ----------------------------------------------------------------------------
#
# Purpose:
#   To download source code and apply patches to enable them for use
#   with the DDK.
#
# Usage:
#   cd driver/product/bldsys
#   ./download_and_patch.sh

# Abort if something seems wrong.
set -e

# Change to product directory if currently elsewhere.
START_DIR=$(pwd)
SCRIPT_DIR=$(dirname $0)
[ -d "$SCRIPT_DIR/.." ] && cd "$SCRIPT_DIR/.."

patchdir()
{
    BASENAME=$(basename "${1}")
    pushd "${1}" >/dev/null
    echo "Patching: $BASENAME"
    patch --quiet -p1 < "../${BASENAME}.patch"
    popd >/dev/null
}

git_download_and_patch()
{
    CHECKOUT_PATH="${1}"
    # Clear anything from previous calls
    unset SHA
    unset URL

    # .url must setup the following environment vars:
    #  URL public git repository location
    #  SHA branch, tag or SHA to checkout
    source "$CHECKOUT_PATH.url"

    if [ -z "$SHA" -o -z "$URL" ] ; then
        echo "ERROR: ${CHECKOUT_PATH}.url does not contain URL and SHA"
        exit 1
    fi

    echo "Downloading: $URL"
    git clone -n "$URL" "$CHECKOUT_PATH"
    git -C "$CHECKOUT_PATH" checkout $SHA
    patchdir "$CHECKOUT_PATH"
}

svn_download_and_patch()
{
    CHECKOUT_PATH="${1}"
    BASENAME=$(basename "${1}")
    URL=$(cat "${CHECKOUT_PATH}.url")
    echo "Downloading: $URL"
    svn export --quiet --non-interactive --trust-server-cert "$URL" "$CHECKOUT_PATH"
    patchdir "$CHECKOUT_PATH"
}

git_download_and_patch "bldsys/bob/blueprint"
#if [ -f "compiler/llvm.url" ] ; then
#    svn_download_and_patch "compiler/llvm"
#fi
#if [ -f "compiler/clang.url" ] ; then
#    svn_download_and_patch "compiler/clang"
#fi
