#!/usr/bin/env python
#
# Copyright:
# ----------------------------------------------------------------------------
# This confidential and proprietary software may be used only as authorized
# by a licensing agreement from ARM Limited.
#      (C) COPYRIGHT 2018 ARM Limited, ALL RIGHTS RESERVED
# The entire notice above must be reproduced on all authorized copies and
# copies may only be made to the extent permitted by a licensing agreement
# from ARM Limited.
# ----------------------------------------------------------------------------
"""Construct DDK data in the format required by the PPA team."""

from __future__ import print_function

import json
import os.path
import subprocess


def generate_ddk_json(config_filename, ddk_filename):
    """Read in the json config data, add ddk info and write out"""
    with open(config_filename) as fp:
        properties = json.load(fp)["Properties"]

    branch = "not a git repository"
    revision = ""
    timestamp = ""
    try:
        with open(os.devnull, 'w') as devnull:
            output = subprocess.check_output(['git', 'rev-parse', '--is-inside-work-tree'], stderr=devnull).strip()

        if output == b'true':
            branch = subprocess.check_output(['git', 'branch', '--contains'])[1:].strip().decode('utf-8')
            revision = subprocess.check_output(['git', 'rev-parse', 'HEAD']).strip().decode('utf-8')
            timestamp = subprocess.check_output(['git', 'log', '-1', '--pretty=format:%ad', revision]).decode('utf-8')
    except subprocess.CalledProcessError:
        # not zero return code indicates that we're not in a git tree
        pass

    ddk = {
        'revision': revision,
        'last_changed_date': timestamp,
        'branch': branch,
        'build_option': properties,
    }

    json_str = json.dumps({"ddk": ddk}, sort_keys=True, indent=4, separators=(',', ': '))

    with open(ddk_filename, 'w') as fp:
        fp.write(json_str)


def main():
    """Console entry point"""
    import argparse
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("config", help="Path to config JSON file")
    parser.add_argument("-o", "--output", default="ddk.json",
                        help="Path to DDK JSON file to be generated")
    args = parser.parse_args()

    generate_ddk_json(args.config, args.output)


if __name__ == "__main__":
    main()
