#!/bin/bash

COMMAND=$1
OPTION=$2
DIR_3D=$3



#================================================================
# you can specify the proper settings based on your environment
#PRJ_NAME=develop.android.magellan.gen.nand
BIFROST_DIR=`pwd`
PRJ_SYSTEM_PATH=$BIFROST_DIR/../../../../..

export ANDROID_SUBDIR_TMP=$(grep ANDROID_SUBDIR $PRJ_SYSTEM_PATH/.config | cut -f2 -d= | cut -f2 -d\")
export ARCH=arm

if [ -f $PRJ_SYSTEM_PATH/../linux/linux-4.4.3/.config ]
then
    export KVER=4.4.3
    export KDIR=$PRJ_SYSTEM_PATH/../linux/linux-4.4.3
elif [ -f $PRJ_SYSTEM_PATH/../linux/linux-4.14/.config ]
then
    export KVER=4.14
	export KDIR=$PRJ_SYSTEM_PATH/../linux/linux-4.14
else
    echo "Cannot find the file of $KDIR/.config !!"
    exit 1
fi

if [ -z $CROSS_COMPILE ]
then
	customer_id=$(grep  ^CUSTOMER_TV00\.=y ../../../../../.config  | cut -f1 -d=)

	if [ "$customer_id" = "CUSTOMER_TV006" ]
	then
		export CROSS_COMPILE=$PRJ_SYSTEM_PATH/tmp/toolchain/arm-lg115x-linux-gnueabi-4.8-2014.11-i686/bin/arm-linux-
		export TOOLCHAIN_PATH=$PRJ_SYSTEM_PATH/tmp/toolchain/arm-lg115x-linux-gnueabi-4.8-2014.11-i686/bin
	else
        if [ "$ANDROID_SUBDIR_TMP" = "oreo" ] && [ "$KVER" = "4.4.3" ]; then
            export CROSS_COMPILE=$PRJ_SYSTEM_PATH/tmp/toolchain/asdk-5.4.1-a53-EL-4.4-g2.23-a32nt-161027/bin/asdk-linux-
		    export TOOLCHAIN_PATH=$PRJ_SYSTEM_PATH/tmp/toolchain/asdk-5.4.1-a53-EL-4.4-g2.23-a32nt-161027/bin
        else
		    export CROSS_COMPILE=$PRJ_SYSTEM_PATH/tmp/toolchain/asdk-6.4.1-a55-EL-4.4-g2.26-a32nut-170810/bin/asdk-linux-
		    export TOOLCHAIN_PATH=$PRJ_SYSTEM_PATH/tmp/toolchain/asdk-6.4.1-a55-EL-4.4-g2.26-a32nut-170810/bin
		    export CC=$PRJ_SYSTEM_PATH/tmp/toolchain/asdk-6.4.1-a55-EL-4.4-g2.26-a32nut-170810/bin/asdk-linux-gcc
		    export CXX=$PRJ_SYSTEM_PATH/tmp/toolchain/asdk-6.4.1-a55-EL-4.4-g2.26-a32nut-170810/bin/asdk-linux-g++
		fi
	fi
fi
export PATH=$TOOLCHAIN_PATH:$PATH
#================================================================
export MYSCONS=$PRJ_SYSTEM_PATH/tools/scons-2.3.0
export SCONS_LIB_DIR=$MYSCONS/engine

SCONS_OPTIONS="v=1 -j8"
PROFILE="profile=kernel-release"
#================================================================

export TARGET_PRODUCT=$(grep ANDROID_TARGET_PRODUCT $PRJ_SYSTEM_PATH/.config | cut -f2 -d= | cut -f2 -d\")
export DISCIMAGE=$ANDROID_ROOT/out/target/product/$TARGET_PRODUCT
export DC_SIZE_720P=$(grep EURASIA_3D_720P $PRJ_SYSTEM_PATH/.config | cut -f2 -d= | cut -f2 -d\")
export DC_SIZE_768P=$(grep EURASIA_3D_768P $PRJ_SYSTEM_PATH/.config | cut -f2 -d= | cut -f2 -d\")
export DC_SIZE_900P=$(grep EURASIA_3D_900P $PRJ_SYSTEM_PATH/.config | cut -f2 -d= | cut -f2 -d\")
export DC_SIZE_1080P=$(grep EURASIA_3D_1080P $PRJ_SYSTEM_PATH/.config | cut -f2 -d= | cut -f2 -d\")
export CONFIG_MALI_FB=y
export ANDROID=y


if [ -z $COMMAND ]
then
	echo "Usage:"
	echo "sh build.sh [option] | [option2] [option3]"
	echo "[option]:"
	echo "	make: make 3D related kernel modules"
	echo "	clean: clean 3D related kernel modules"
	echo "	install: install 3D related kernel modules"
	echo "  prebuilt: pack prebuilt tgz"
	echo "[option2] ONLY for make and install"
	echo "	debug: install 3D related kernel modiles from binary_bifrost_android_debug/"
	echo "	release: install 3D related kernel modiles from binary_bifrost_android_release/"
	echo "[option3] ONLY for prebuilt"
	echo "  directoryName of this ddk. ex: sh build.sh prebuilt release 3d_graphic_android_r11p0-01rel0"
else

	if [ -z $OPTION ]
	then
    	INSTALL_FOLDER=binary_bifrost_android_release
		PROFILE="profile=kernel-release"
	else
		if [ $OPTION = "debug" ]
		then
			INSTALL_FOLDER=binary_bifrost_android_debug
			PROFILE="profile=kernel-debug"
		else
			INSTALL_FOLDER=binary_bifrost_android_release
			PROFILE="profile=kernel-release"
		fi
	fi


	if [ $COMMAND = "clean" ]
	then
		if [ -z $OPTION ]
        then
        	OPTION="release"
		fi

        python $MYSCONS/script/scons -f bldsys/sconstruct $PROFILE $SCONS_OPTIONS --clean
        echo "rm -r $INSTALL_FOLDER;"
		rm -r $INSTALL_FOLDER;
        echo "rm -r build;"
		rm -r build/;
        echo "rm -r bldsys/*.pyc;"
		rm -r bldsys/*.pyc;
		#echo "rm -r $BIFROST_DIR/kernel/drivers/video/mali_fb;"
		#rm -rf $BIFROST_DIR/kernel/drivers/video/mali_fb;
	fi

	if [ $COMMAND = "make" ]
    then
		if [ -z $OPTION ]
		then
			export BUILD=release
		else
			if [ $OPTION = "debug" ]
			then
				export BUILD=debug
			else
				export BUILD=release
			fi
		fi
		echo "mali km build: python $MYSCONS/script/scons -f bldsys/sconstruct $PROFILE $SCONS_OPTIONS progs_install=$INSTALL_FOLDER"
        python $MYSCONS/script/scons -f bldsys/sconstruct $PROFILE $SCONS_OPTIONS progs_install=$INSTALL_FOLDER
		#scons -f bldsys/sconstruct $PROFILE $SCONS_OPTIONS 
		#cp gator.ko ./$INSTALL_FOLDER/
        exit 0
    fi

	if [ $COMMAND = "install" ]
    then
		cd ./$INSTALL_FOLDER/
		echo "Install completed!"
		exit 0
    fi
    
	if [ $COMMAND = "prebuilt" ]
    then
		if [ -z $DIR_3D ]
		then
			echo " [option3] for prebuilt not set!! "
			exit -1
		fi
		exit 0
	fi    
fi
